#include "math.h"
#include "stdlib.h"
#include "stdio.h"


#define USIZE sizeof(unsigned)
#define BUFFER_SIZE 100
#define BYTES_TO_PRINT 7

typedef unsigned char uchar;


__inline__ uint log2(uint n)
{	
  uint i=0;
  n >>= 1;

  for(; n != 0;   n >>= 1, i++) ;
  return(i);
}


/* ==================================================================
   This procedure takes the pointer B to the first byte of the buffer
   and cancels BUFFER_LEN bytes of it.
   ================================================================== */

void clean_buffer(uchar *buffer, int buffer_len)
{	
  int i;
 
  for(i=0; i< buffer_len; i++)
    buffer[i] = 0;
}


/* ==================================================================
   This procedure takes the pointer B to the first byte of the buffer
   where we have to write and the #bits to skip (*BIT_OFFSET) and
   stores the least significant BIT_LEN bits in SOURCE. The procedure
   also updates the buffer pointer and the offset.
   ================================================================== */

void write_buffer(uchar **b, int *bit_offset, uint source, int bit_len)
{								      
  int writeable = bit_len + (*bit_offset); // fictitious heading bits
  int right_shift, left_shift;

  if(bit_len == 0) { return; }

  if(writeable <= 8){ // write on one byte only
    (**b) |= (source << (8-writeable)); 
    *bit_offset = writeable % 8;
    if(*bit_offset == 0) (*b)++;

  } else { 


      do {
	
	right_shift = writeable-8;  // to take one byte
	(**b) |= (source >> right_shift); // takes the leftward byte
	
	left_shift = 8*USIZE - right_shift; // to cancel the written bits
	source = (source << left_shift) >> left_shift; // cancels written bits
	
	(*b)++; 
	writeable -= 8;        

      } while(writeable > 8);  // more than one byte to write
      
      (**b) |= (source << (8-writeable)); // the remaining bits 
      (*bit_offset) = ((*bit_offset) + bit_len) % 8;
      if(*bit_offset == 0) (*b)++; 
  }
}


/* ==================================================================
   This procedure takes the pointer B to the first byte of the buffer
   where we have to read and the #bits to skip (*BIT_OFFSET) and
   computes the number of 0s following. The procedure also
   updates the buffer pointer and the offset.
   ================================================================== */
int read_zero_number(uchar **b, int *bit_offset)
{								      
  uchar byte = ((**b)<< (*bit_offset)); // delete offset #bits
  int len=0;


  if(!byte) { (*b)++; byte = **b; len = 8- (*bit_offset);  }

  while(!byte){
    len += 8;
    (*b)++;
    byte = **b;
  }

  while((byte & 0x80) != 128){
    byte <<= 1;
    len++;
  }
  
  (*bit_offset) = ((*bit_offset) + len) % 8;
  return(len);
}


/* ==================================================================
   This procedure takes the pointer B to the first byte of the buffer
   where we have to read and the #bits to skip (*BIT_OFFSET) and reads
   in LEN bits. The procedure also updates the buffer pointer and the
   offset.
   ================================================================== */

uint read_value(uchar **buffer, int *bit_offset, int len)
{								      
  uchar **b=buffer;
  int i=0;
  uint value=0;
  int remaining;

  if (len==0) { return(0); }

  value = ((**b) & (0xFF >> (*bit_offset)));
  if ((*bit_offset)+len <=8) 
    {
      value >>= (8- (*bit_offset+len));
      (*bit_offset) = ((*bit_offset)+len) % 8;
      if ((*bit_offset) == 0) (*b)++;
      return(value);
    }

  i = (8- (*bit_offset));
  (*b)++;

  for(;(i+8)<len;i+=8)
    {
      value = (value<<8) | (**b);
      (*b)++;
    }

  remaining = len-i;
  value <<= remaining;

  value |= ((**b) >> (8-remaining));  // take the last bits
  (*bit_offset) = ((*bit_offset) + len) % 8;
  if(*bit_offset == 0) (*b)++;

  return(value);
}


/* ==================================================================
   GAMMA CODING, with updating of the buffer and offset pointers.
   ================================================================== */

void write_gamma_code(uchar **buffer, int *bit_offset, uint n)
{
  int len = log2(n);

  write_buffer(buffer,bit_offset,0,len); // possibly len=0
  write_buffer(buffer,bit_offset,n,len+1);
}
  
uint read_gamma_code(uchar **buffer, int *bit_offset)
{

  int len;
  uint value;

  len = read_zero_number(buffer,bit_offset);
  value = read_value(buffer,bit_offset,len+1);

  return(value);
}



/* ==================================================================
   DELTA CODING, with updating of the buffer and offset pointers.
   ================================================================== */

void write_delta_code(uchar **buffer, int *bit_offset, uint n)
{
  uint len = (uint) log2(n);
  uint pow2 = (uint) pow(2, len);

  write_gamma_code(buffer,bit_offset,len+1);
  write_buffer(buffer,bit_offset,n-pow2,len);
}
  
uint read_delta_code(uchar **buffer, int *bit_offset)
{

  uint len;
  uint value;
  uint pow2;

  len = read_gamma_code(buffer,bit_offset);
  len = len - 1;
  value = read_value(buffer,bit_offset,len);
  pow2  = (uint) pow(2, len);

  return(value + pow2);
}


/* ==================================================================
   GAMMA-GOLOMB CODING, it updates the buffer and offset pointers.
   ================================================================== */

void write_gamma_golomb_code(uchar **buffer, int *bit_offset, uint n, uint k)
{
  uint q = floor((n-1)/k);
  uint r = n - q*k - 1;
  uint log2_k = (uint) log2(k);
  uint pow2_k = (uint) pow(2, log2_k);
  uint d = 2*pow2_k - k;

  write_gamma_code(buffer,bit_offset,q+1);

  if (r < d)
    write_buffer(buffer,bit_offset,r,log2_k);
  else
    write_buffer(buffer,bit_offset,r+d,log2_k+1);
}
  
uint read_gamma_golomb_code(uchar **buffer, int *bit_offset, uint k)
{
  uint q,r;
  uint log2_k = (uint) log2(k);
  uint pow2_k = (uint) pow(2, log2_k);
  uint d = 2*pow2_k - k;

  q = read_gamma_code(buffer,bit_offset)-1;
  r = read_value(buffer,bit_offset,log2_k);

  if (r < d)
    return(r+1+q*k);
  else
    {  // fetches one more bit
      r <<= 1;
      r |= read_value(buffer,bit_offset,1);
      return(r+1+q*k-d);
    }
}



/* ==================================================================
   GOLOMB CODING, it updates the buffer and offset pointers.
   ================================================================== */

void write_golomb_code(uchar **buffer, int *bit_offset, uint n, uint k)
{
  uint q = floor((n-1)/k);
  uint r = n - q*k - 1;
  uint log2_k = (uint) log2(k);
  uint pow2_k = (uint) pow(2, log2_k);
  uint d = 2*pow2_k - k;
  int i;
  
  // write the long sequences of zeros in groups of 32 bits
  for(i=0; i< floor(q/32); i++)
    write_buffer(buffer, bit_offset, (uint) 0, 32);
  write_buffer(buffer, bit_offset, (uint) 0, q % 32);
  
  // write the separating 1
  write_buffer(buffer, bit_offset, 1, 1);

  if (r < d)
    write_buffer(buffer,bit_offset,r,log2_k);
  else
    write_buffer(buffer,bit_offset,r+d,log2_k+1);
}


  
uint read_golomb_code(uchar **buffer, int *bit_offset, uint k)
{
  uint q,r;
  uint log2_k = (uint) log2(k);
  uint pow2_k = (uint) pow(2, log2_k);
  uint d = 2*pow2_k - k;

  q = read_zero_number(buffer,bit_offset);
  read_value(buffer,bit_offset,1); // skip the separating 1
  r = read_value(buffer,bit_offset,log2_k);

  if (r < d)
    return(r+1+q*k);
  else
    {  // fetches one more bit
      r <<= 1;
      r |= read_value(buffer,bit_offset,1);
      return(r+1+q*k-d);
    }
}



